import numpy as np
import matplotlib.pyplot as plt
plt.ion()
plt.show()


import torch         # GPU + autodiff library
from   torch.autograd import Variable, grad

# Set the parameters of our model:
g      = torch.Tensor( [ 9.81] )
m      = torch.Tensor( [ 15. ] )
source = torch.Tensor( [0.,.5] )
target = torch.Tensor( [7.,2.] )

# Wrap them into "autodiff" graph nodes:
g      = Variable(      g, requires_grad = True )
m      = Variable(      m, requires_grad = True )
source = Variable( source, requires_grad = True )
target = Variable( target, requires_grad = True )

def cost(m, g, P, windy=False) :
  "Cost associated to a simple ballistic problem."

  def Emec(q,p) :
    "Particle of mass m in a gravitational field g."
    return m*g*q[1] + (p**2).sum() / (2*m)

  qt = source ; pt = P
  qts = [qt]

  for it in range(10) : # Simple Euler scheme
    [dq,dp] = grad(Emec(qt,pt), [qt,pt], create_graph=True)
    if windy :
      dq += qt[1] * 20 * Variable(torch.randn(2))

    qt = qt + .1 * dp
    pt = pt - .1 * dq
    qts.append(qt)
  # Return the squared cost $|q_1 - target|^2_2$
  return ((qt - target)**2).sum(), qts

P = torch.Tensor( [60., 30.] )
P = Variable( P, requires_grad = True )

fig = plt.figure(figsize=(10,5))
ax = fig.add_subplot(111, aspect='equal')

GRAV = True
WIND = True

lr = .1 # .1 if GRAV
for it in range(101) :
  C, qts = cost(m,g,P, WIND)
  print("Cost : ", C.data.cpu().numpy()[0])
  if GRAV :
    dg = grad(C, [g])[0]
    g.data -= lr*dg.data
  else :
    dP = grad(C, [P])[0]
    P.data -= lr*dP.data

  if it % 10 == 0 :
    ax.clear()
    a = P.data.cpu().numpy()
    a = np.arctan2(a[0], a[1])
    
    qts = torch.stack(qts).data.cpu().numpy()
    ax.plot(qts[:,0], qts[:,1], '-o', color='#0051FF')

    if WIND :
      for _ in range(10) :
        _, qts = cost(m,g,P, WIND)
        qts = torch.stack(qts).data.cpu().numpy()
        ax.plot(qts[:,0], qts[:,1], '-o', color='#88B0FF', zorder=1.5)

    circle1 = plt.Circle((7, 2), .8, color='#FF8A8A')
    circle2 = plt.Circle((7, 2), .6, color='#FFFFFF')
    circle3 = plt.Circle((7, 2), .4, color='#FF8A8A')
    circle4 = plt.Circle((7, 2), .2, color='#FFFFFF')

    ax.add_artist(circle1)
    ax.add_artist(circle2)
    ax.add_artist(circle3)
    ax.add_artist(circle4)

    ax.set_yticklabels([])
    ax.set_xticklabels([])
    ax.xaxis.set_ticks_position('none') 
    ax.yaxis.set_ticks_position('none') 
    plt.ylim( 0,5)
    plt.xlim(-2,8)
    plt.gca().set_aspect('equal', adjustable='box')
    plt.draw()
    if   GRAV == True : suff = "grav"
    elif WIND == True : suff = "windy"
    else :              suff = "simple"
    plt.savefig("output/OC_"+suff+"_"+str(it)+".png")

plt.show(block=True)